<?xml version="1.0"?>
<xsl:transform xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
               xmlns:stream="http://www.data-structures.org/stream/" 
               xmlns:xs="http://www.w3.org/2001/XMLSchema"
               version="2.0">
    
    <!--
        Library of Stream Functions
        Author: Roger L. Costello
        Date: September 11, 2010  
        Version 1.0      
     -->
    
    <xsl:include href="stream-utilities.xsl"/>
    
    <!-- 
        Here are the stream functions:
        
        stream:empty(stream): returns true if the stream is empty, false otherwise.
        
        stream:first(stream): returns the first value in the stream.
        
        stream:rest(stream): recall that a stream contains a first value and a promise to evaluate the rest if requested. Invoke this function
        when you want to cash in on that promise. Thus, this function evaluates the rest of the stream. Actually, it returns the first value in 
        the rest and a promise to evaluate the rest if requested.
        
        stream:ref(stream, i): returns the i'th value in the stream. Example: if the stream is the interval 1 to 10 then stream:ref(stream, 2) returns 2.
        
        stream:iterator(stream, N): returns the first N values in the stream.
        
        stream:iterator(stream): returns all the values in the stream (it keeps recursively processing the stream until it reaches an empty stream).
        
        stream:map(proc, stream): applies (i.e., maps) the procedure, proc, to each value in the stream. Actually, if applies proc to the first value
        in the stream and returns it along with a promise to apply proc the rest if requested.
        
        stream:enumerate-interval(low, high): returns the integers from low to high. Actually, it returns the first integer and a promise to enumerate 
        the rest if requested.
        
        stream:enumerate-following-siblings(node): returns the (element) node and all its following siblings. Actually, it returns the first node and a 
        promise to provide the rest if requested.
        
        stream:filter(predicate, stream): this applies predicate to each value in the stream. Actually, it applies the predicate to the first value 
        in the stream and a promise to apply the predicate to the rest if requested. Example: suppose the stream is 1 to 10 and the predicate is odd
        (is the number an odd number). Then stream:filter(odd, stream) returns 1, 3, 5, 7, 9.
        
        stream:accumulate(op, initial, stream): this applies the operation, op, to the stream using "initial" as the starting value. Actually, it applies
        op to the first value in the stream and a promise to apply it to the rest if requested. For example, suppose op is multiply, the initial value is 1,
        and the stream is the integers from 1 to 10. Then stream:accumulate(multiply, 1, stream) produces this stream: 1 * 2 * 3 * 4 * ... * 10.
        
        stream:abs(stream): this returns the absolute value of each value in the stream. Actually, it returns the absolute value of the first value in the
        stream and a promise to return the absolute value of the rest if requested.
        
        stream:credit-debit(balance, stream): this is used exclusively with an XML document containing a sequence of sibling <withdraw> and <deposit> elements. 
        It goes through each node in the sequence and credits/debits the balance. Actually, it credits/debits the first node along with a promise to do the
        rest if requested.
        
    -->
    
    <!-- 
        
        For some of the functions there is also a partner template rule. For example, there is a function, stream:enuerate-interval, and a template that 
        matches stream:enumerate-interval. In addition, there is a <stream:enumerate-interval /> element and a variable, enumerate-interval, which 
        references that element. What's this for? Recall that each stream function returns the first value and a promise to return the rest if requested. 
        For example, stream:enumerate-interval(10000, 100000000) returns this: (10000, <stream:delay/>, <stream:enumerate-interval/>, 10001, 100000000).
        10000 is the first value in the interval and the remaining represents the promise the return the rest of the interval if requested. Suppose it is
        requested. Then the software must extract from the stream the name of the function to evaluate. The name is indicated by the <stream:enumerate-interval/>
        element. Hold on for this, it is really cool: To evaluate the function, use xsl:apply templates with select equal to the <stream:enumerate-interval/>
        element. The force function (in stream-utilities.xsl) does this. Effectively, what is being done is the name of a function is passed to the force()
        function and it invokes the function. I wish that I could take credit for this design pattern, but the credit goes to Dmitre Novatchev. It is his
        genius that created this.
        
    -->
    
    <!-- 
        
        September 12, 2010
        
       ac on the XSL list noted that these:
       
             <xsl:variable name="map" select="document('')//stream:map" />
       
             <stream:map/>
       
       could be replaced with just this:
       
             <xsl:variable name="map" as="element()"><stream:map/></xsl:variable>
       
       Ditto for all the other places.
       
       So I made the changes. Thanks ac!
       
    -->
    
    <xsl:function name="stream:empty" as="xs:boolean">
        <xsl:param name="stream" as="item()*"/>
        
        <xsl:sequence select="xs:untypedAtomic($stream[1]) eq $the-empty-stream" />
        
    </xsl:function>
    

    <xsl:function name="stream:first" as="item()+">
        <xsl:param name="stream" as="item()*"/>
        
        <xsl:choose>
            <xsl:when test="stream:empty($stream)">
                <xsl:sequence select="$the-empty-stream" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:sequence select="$stream[1]" />
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    

    <xsl:function name="stream:rest" as="item()*">
        <xsl:param name="stream" as="item()*"/>
        
        <xsl:choose>
            <xsl:when test="stream:empty($stream)">
                <xsl:sequence select="$the-empty-stream" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:sequence select="stream:force($stream[3], $stream)" />
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    
    
    <xsl:function name="stream:ref" as="item()*">
        <xsl:param name="stream" as="item()*"/>
        <xsl:param name="index" as="xs:integer" />
        
        <xsl:choose>
            <xsl:when test="$index eq 1">
                <xsl:sequence select="stream:first($stream)" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:sequence select="stream:ref(stream:rest($stream), $index - 1)" />
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    
    
    <xsl:function name="stream:iterator" as="item()*">
        <xsl:param name="stream" as="item()*"/>
        <xsl:param name="count" as="xs:integer"/>
        
        <xsl:choose>
            <xsl:when test="$count eq 0" />
            <xsl:otherwise>
                <xsl:sequence select="stream:first($stream)" />
                <xsl:sequence select="stream:iterator(stream:rest($stream), $count - 1)" />
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    
    
    <xsl:function name="stream:iterator" as="item()*">
        <xsl:param name="stream" as="item()*"/>
        
        <xsl:choose>
            <xsl:when test="stream:empty($stream)" />
            <xsl:otherwise>
                <xsl:sequence select="stream:first($stream)" />
                <xsl:text> </xsl:text>
                <xsl:sequence select="stream:iterator(stream:rest($stream))" />
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    
    
    <xsl:variable name="map" as="element()"><stream:map /></xsl:variable>
    
    <xsl:template match="stream:map" as="item()*">
        <xsl:param name="proc" select="()" as="element()?" />
        <xsl:param name="stream" as="item()*" />
        
        <xsl:variable name="proc-actual" select="if (empty($proc)) then $stream[4] else $proc" />
        <xsl:variable name="stream-actual" select="if (empty($stream)) then $stream else $stream" />
        
        <xsl:sequence select="stream:map($proc-actual, $stream-actual)" />
        
    </xsl:template>
    
    
    <xsl:function name="stream:map" as="item()*">
        <xsl:param name="proc" as="element()" />
        <xsl:param name="stream" as="item()*" />
        
        <xsl:apply-templates select="$proc">
            <xsl:with-param name="stream" select="$stream" />
        </xsl:apply-templates>
        
    </xsl:function>
    

    <xsl:variable name="enumerate-interval" as="element()"><stream:enumerate-interval /></xsl:variable>
    
    <xsl:template match="stream:enumerate-interval" as="item()*">
        <xsl:param name="low" select="()" as="xs:integer?"/>
        <xsl:param name="high" select="()" as="xs:integer?"/>
        <xsl:param name="stream" as="item()*" />
        
        <xsl:variable name="low-actual" select="if (empty($low)) then $stream[4] else $low" />
        <xsl:variable name="high-actual" select="if (empty($high)) then $stream[5] else $high" />
        
        <xsl:sequence select="stream:enumerate-interval($low-actual, $high-actual)" />
        
    </xsl:template>
    
    <xsl:function name="stream:enumerate-interval" as="item()*">
       <xsl:param name="low" as="xs:integer"/>
       <xsl:param name="high" as="xs:integer"/>
       
        <xsl:choose>
            <xsl:when test="$low gt $high">
                <xsl:sequence select="$the-empty-stream" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:sequence select="stream:construct($low, ($enumerate-interval, $low + 1, $high))" />
            </xsl:otherwise>
        </xsl:choose>
       
    </xsl:function>
    
    
    <xsl:variable name="enumerate-following-siblings" as="element()"><stream:enumerate-following-siblings /></xsl:variable>
    
    <xsl:template match="stream:enumerate-following-siblings" as="item()*">
        <xsl:param name="node" select="()" as="element()?"/>
        <xsl:param name="stream" as="item()*" />
        
        <xsl:variable name="node-actual" select="if (empty($node)) then $stream[4] else $node" />
        
        <xsl:sequence select="stream:enumerate-following-siblings($node-actual)" />
        
    </xsl:template>
    
    <xsl:function name="stream:enumerate-following-siblings" as="item()*">
        <xsl:param name="node" as="element()?"/>
        
        <xsl:choose>
            <xsl:when test="empty($node)">
                <xsl:sequence select="$the-empty-stream" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:sequence select="stream:construct($node, ($enumerate-following-siblings, $node/following-sibling::*[1]))" />
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    

    <xsl:variable name="filter" as="element()"><stream:filter /></xsl:variable>
    
    <xsl:template match="stream:filter" as="item()*">
        <xsl:param name="predicate" select="()" as="element()?" />
        <xsl:param name="stream" as="item()*" />
        
        <xsl:variable name="predicate-actual" select="if (empty($predicate)) then $stream[4] else $predicate" />
        <xsl:variable name="stream-actual" select="if (empty($predicate)) then $stream[position() ge 5] else $stream" />
        
        <xsl:sequence select="stream:filter($predicate-actual, $stream-actual)" />
        
    </xsl:template>
    
    <xsl:function name="stream:filter" as="item()*">
        <xsl:param name="predicate" as="element()" />
        <xsl:param name="stream" as="item()*" />
        
        <xsl:choose>
            <xsl:when test="stream:empty($stream)">
                <xsl:sequence select="$the-empty-stream" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:variable name="accept-item1" as="xs:boolean">
                    <xsl:apply-templates select="$predicate">
                        <xsl:with-param name="n" select="stream:first($stream)" />
                    </xsl:apply-templates>
                </xsl:variable>
                <xsl:choose>
                    <xsl:when test="$accept-item1">
                        <xsl:sequence select="stream:construct(stream:first($stream), ($filter, $predicate, stream:rest($stream)))" />
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:sequence select="stream:filter($predicate, stream:rest($stream))" />
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    

    <xsl:variable name="stream-accumulate" as="element()"><stream:accumulate /></xsl:variable>
    
    <xsl:template match="stream:accumulate" as="item()*">
        <xsl:param name="op" select="()" as="element()?" />
        <xsl:param name="initial" select="()" as="item()?" />
        <xsl:param name="stream" as="item()*" />
        
        <xsl:variable name="op-actual" select="if (empty($op)) then $stream[4] else $op" />
        <xsl:variable name="initial-actual" select="if (empty($initial)) then $stream[1] else $initial" />
        <xsl:variable name="stream-actual" select="if (empty($op) and empty($initial)) then $stream[position() ge 5] else $stream" />
        
        <xsl:sequence select="stream:accumulate($op-actual, $initial-actual, $stream-actual)" />
        
    </xsl:template>
    
    <xsl:function name="stream:accumulate" as="item()*">
        <xsl:param name="op" as="element()" />
        <xsl:param name="initial" as="item()" />
        <xsl:param name="stream" as="item()*" />
        
        <xsl:choose>
            <xsl:when test="stream:empty($stream)">
                <xsl:sequence select="$the-empty-stream" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:variable name="new-initial">
                    <xsl:apply-templates select="$op">
                        <xsl:with-param name="opnd1" select="$initial" />
                        <xsl:with-param name="opnd2" select="stream:first($stream)" />
                    </xsl:apply-templates>
                </xsl:variable>
                <xsl:sequence select="stream:construct($new-initial, ($stream-accumulate, $op, stream:rest($stream)))" />
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    

    <xsl:variable name="stream-abs" as="element()"><stream:abs /></xsl:variable>
   
    <xsl:template match="stream:abs" as="item()*">
        <xsl:param name="stream" as="item()*" />
                
        <xsl:variable name="stream-actual" select="if (node-name($stream[3]) eq node-name($stream-abs)) then $stream[position() ge 4] else $stream" />
        
        <xsl:sequence select="stream:abs($stream-actual)" />
        
    </xsl:template>
    
    <xsl:function name="stream:abs" as="item()*">
        <xsl:param name="stream" as="item()*" />
        
        <xsl:choose>
            <xsl:when test="stream:empty($stream)">
                <xsl:sequence select="$the-empty-stream" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:choose>
                    <xsl:when test="xs:integer(stream:first($stream)) lt 0">
                        <xsl:sequence select="stream:construct(0 - xs:integer(stream:first($stream)), ($stream-abs, stream:rest($stream)))" />
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:sequence select="stream:construct(xs:integer(stream:first($stream)), ($stream-abs, stream:rest($stream)))" />
                    </xsl:otherwise>
                </xsl:choose>
                
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    
    
    <xsl:variable name="stream-credit-debit" as="element()"><stream:credit-debit /></xsl:variable>
    
    <xsl:template match="stream:credit-debit" as="item()*">
        <xsl:param name="balance" select="()" as="item()?" />
        <xsl:param name="stream" as="item()*" />
        
        <xsl:variable name="balance-actual" select="if (empty($balance)) then $stream[1] else $balance" />
        <xsl:variable name="stream-actual" select="if (empty($balance)) then $stream[position() ge 4] else $stream" />
        
        <xsl:sequence select="stream:credit-debit($balance-actual, $stream-actual)" />
        
    </xsl:template>
    
    <xsl:function name="stream:credit-debit" as="item()*">
        <xsl:param name="balance" as="item()" />
        <xsl:param name="stream" as="item()*" />
        
        <xsl:choose>
            <xsl:when test="stream:empty($stream)">
                <xsl:sequence select="$the-empty-stream" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:variable name="new-balance">
                    <xsl:choose>
                        <xsl:when test="stream:first($stream)[self::deposit]">
                            <xsl:sequence select="$balance + xs:integer(stream:first($stream)/text())" />
                        </xsl:when>
                        <xsl:when test="stream:first($stream)[self::withdrawal]">
                            <xsl:sequence select="$balance - xs:integer(stream:first($stream)/text())" />
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:sequence select="$balance" />
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:variable>
                <xsl:sequence select="stream:construct($new-balance, ($stream-credit-debit, stream:rest($stream)))" />
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>


</xsl:transform>